import _        from 'lodash';
import Prism    from 'prismjs';
import Beautify from 'js-beautify';




// ------------------------------------------------------------------------------------------------
// Global functions
// ------------------------------------------------------------------------------------------------
/**
 * Generate code samples from content.
 *
 * @author Curtis Blackwell
 * @return {void}
 */
window.generateCodeSamples = () => {
  _.forEach($$('.sample'), (sample, i) => {
    let specimen =
      '<div class="specimen">' +
        '<div class="card">' +
          '<div class="card-content">' +
            sample.innerHTML +
          '</div>' +
        '</div>' +
      '</div>';

    let codeSample = sample.className.indexOf('js-vue') > -1
      ? getCodeSample(sample.innerHTML, 'vue-sample-' + i)
      : getCodeSample(sample.innerHTML);

    sample.innerHTML = specimen + '<div class="dna">' + codeSample + '</div>';
  });
};




/**
 * Generate code samples for rendered Vue components.
 *
 * @author Curtis Blackwell
 * @return {void}
 */
window.generateCodeSamplesForRenderedVueComponents = () => {
  _.forEach($$('.sample.js-vue .dna .tab-panels'), (sample, i) => {
    sample.querySelector('li:nth-child(2)').innerHTML = getCodeSample(
      sample.parentNode.previousSibling.querySelector('.card-content').innerHTML
    );
  });
};




/**
 * Generate nav from headings in content.
 *
 * @author Curtis Blackwell
 * @return {void} [description]
 */
window.generateNav = (sidebar) => {
  let navItems       = '';
  let previousHLevel = false;
  let idMap          = [];

  _.forEach($$('.js-target'), (h) => {
    // Get the level of the heading (h2 -> 2).
    let level    = parseInt(h.tagName.substring(1));
    // Set the index at this level.
    idMap[level] = _.kebabCase(h.textContent);
    let id       = getHeadingId(idMap, level);

    // Set the `id` on the heading so we can link to it.
    h.id      = id;
    navItems += getNavItem(level, previousHLevel, '<a href="#' + id + '">' + h.textContent + '</a>');

    previousHLevel = level;
  });
  // Close the last nav item's `li`.
  navItems += '</li>';

  sidebar.innerHTML = '<ol>' + navItems + '</ol>';
};




/**
 * Init method for sticky elements.
 *
 * Since it uses the `window.onscroll` event, only one thing can be sticky.
 *
 * When the user scrolls past the `$oov`, the `$sticky` has `.is-sticky` added. `.is-sticky` gets
 * removed when the `$oov` comes back in view.
 *
 * @author Curtis Blackwell
 * @param  {object} $oov    Out of viewport element.
 * @param  {object} $sticky Sticky element.
 * @return {void}
 */
window.letsGetSticky = ($oov, $sticky) => {
  window.onscroll = () => {
    if ($oov.getBoundingClientRect().top < -$oov.offsetHeight) {
      $sticky.classList.add('is-sticky');
    } else {
      $sticky.classList.remove('is-sticky');
    }
  };
};




// ------------------------------------------------------------------------------------------------
// Helpers
// ------------------------------------------------------------------------------------------------
/**
 * Format HTML.
 *
 * @author Curtis Blackwell
 * @param  {string} html
 * @return {string}
 */
let formatHTML = (html) => {
  return Beautify.html(html, {
    indent_size:                 2,
    indent_char:                 ' ',
    indent_with_tabs:            false,
    eol:                         '\\n',
    end_with_newline:            false,
    preserve_newlines:           true,
    max_preserve_newlines:       4,
    indent_inner_html:           false,
    brace_style:                 'collapse-preserve-inline',
    indent_scripts:              'normal',
    wrap_line_length:            80,
    wrap_attributes:             'auto',
    wrap_attributes_indent_size: 2,
    unformatted:                 false,
    extra_liners:                false,
  });
};




/**
 * Prepare HTML to be displayed rather than rendered.
 *
 * @author Curtis Blackwell
 * @param  {string} html
 * @return {string}
 */
let prepHTMLForDisplay = (html) => {
  return _.chain(formatHTML(html))
          // Replace `<` with HTML entity to prevent rendering as DOM nodes.
          .replace(/</g, '&lt;')
          // Insert a zero-width space between `{{` to prevent Vue variables from parsing.
          .replace(/\{\{/g, '{&zwj;{')
          // Remove cruft added by js-beautify to boolean attributes (and some Vue directives).
          .replace('=""', '')
          .value();
};




/**
 * Get a code sample from HTML.
 *
 * @author Curtis Blackwell
 * @param  {string} html
 * @param  {string} refName Name to use as ref for tabs. Must be unique.
 * @return {string}         Markup wrapped up appropriately.
 */
let getCodeSample = (html, refName) => {
  let codeSample =
    '<pre class="card">' +
      '<code class="language-html">' + prepHTMLForDisplay(html) + '</code>' +
    '</pre>';

  if (refName) {
    codeSample =
      '<hex-tabs target="' + refName + '">' +
        '<hex-tab>Vue</hex-tab>' +
        '<hex-tab>HTML</hex-tab>' +
      '</hex-tabs>' +

      '<hex-tab-panels ref="' + refName + '">' +
        '<hex-tab-panel>' + codeSample + '</hex-tab-panel>' +
        '<hex-tab-panel></hex-tab-panel>' +
      '</hex-tab-panels>';
  }

  return codeSample;
};




/**
 * Figure out the appropriate `id` for a given heading.
 *
 * @author Curtis Blackwell
 * @param  {object} idMap Map of `id`s to allow deeper-level headings to grab higher-level headings'
 *                        `id`s, which helps prevent duplicate `id`s when headings match.
 *
 *                        Ex. This map:
 *                        ```js
 *                        let idMap = [
 *                          2: 'colors',
 *                          3: 'background',
 *                          4: 'base',
 *                        ];
 *                        ```
 *
 *                        Results in these headings:
 *                        h2#colors                 Colors
 *                        h3#colors-background      Background
 *                        h4#colors-background-base Base
 *
 * @param  {number} level Heading level (h2 -> 2).
 * @return {string}
 */
let getHeadingId = (idMap, level) => {
  // Concatenate the value at the level's index and each one before it.
  return _.chain(_.slice(idMap, 0, level + 1))
          .join('-')
          .replace(/^[^a-z]*/, '')
          .value();
};




/**
 * Get a nav item.
 *
 * Each heading's `li` gets left open. The following heading closes it. (The last heading gets
 * closed after the loop that calls this function.)
 *
 * @author Curtis Blackwell
 * @param  {number} level          Heading level (h2 -> 2).
 * @param  {number} previousHLevel Previous heading's level.
 * @param  {string} link           Markup for the link.
 * @return {string}
 */
let getNavItem = (level, previousHLevel, link) => {
  if (!previousHLevel) {
    // First heading.
    return '<li>' + link;

  } else if (level === previousHLevel) {
    // If the heading is at the same level as the previous heading...
    return '</li><li>' + link;

  } else if (level > previousHLevel) {
    // If the heading is a level deeper than the previous heading. (Jumping more than one level
    // is wrong.)
    return '<ol><li>' + link;

  } else if (level < previousHLevel) {
    // If the heading is a level higher than the previous heading...
    return '' +
      // Close an `li` and `ol` for each level above. For example, if the last heading was an `h4`
      // and this one is an `h2`, close 2 `li`s and `ol`s.
      _.times(previousHLevel - level, () => {
        return '</li></ol>';
      }).join('') +
      // ...then close the previous heading's `li` and carry on, my wayward son.
      '</li><li>' + link;
  }
};




// ------------------------------------------------------------------------------------------------
// Convenience functions
// ------------------------------------------------------------------------------------------------
window.$ = (selector) => {
  return document.querySelector(selector);
};

window.$$ = (selector) => {
  return document.querySelectorAll(selector);
};
